/*
 * linux/arch/arm/mach-ep93xx/gpio.c
 *
 * Generic EP93xx GPIO handling
 *
 * Copyright (c) 2008 Ryan Mallon <ryan@bluewatersys.com>
 *
 * Based on code originally from:
 *  linux/arch/arm/mach-ep93xx/core.c
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/seq_file.h>

#include <asm/arch/ep93xx-regs.h>
#include <asm/io.h>
#include <asm/gpio.h>

struct ep93xx_gpio_chip {
	struct gpio_chip	chip;

	unsigned int		data_reg;
	unsigned int		data_dir_reg;
	unsigned int		debounce_reg;
};

#define to_ep93xx_gpio_chip(c) container_of(c, struct ep93xx_gpio_chip, chip)

/* From core.c */
extern void ep93xx_gpio_int_mask(unsigned line);
extern void ep93xx_gpio_update_int_params(unsigned port);

static int ep93xx_gpio_direction_input(struct gpio_chip *chip, unsigned offset)
{
	struct ep93xx_gpio_chip *ep93xx_chip = to_ep93xx_gpio_chip(chip);
	unsigned long flags;
	u8 v;

	local_irq_save(flags);
	v = __raw_readb(ep93xx_chip->data_dir_reg);
	v &= ~(1 << offset);
	__raw_writeb(v, ep93xx_chip->data_dir_reg);
	local_irq_restore(flags);

	return 0;
}

static int ep93xx_gpio_direction_output(struct gpio_chip *chip,
					unsigned offset, int val)
{
	struct ep93xx_gpio_chip *ep93xx_chip = to_ep93xx_gpio_chip(chip);
	unsigned long flags;
	int line;
	u8 v;

	local_irq_save(flags);

	/* Set the value */
	v = __raw_readb(ep93xx_chip->data_reg);
	if (val)
		v |= (1 << offset);
	else
		v &= ~(1 << offset);
	__raw_writeb(v, ep93xx_chip->data_reg);

	/* Drive as an output */
	line = chip->base + offset;
	if (line <= EP93XX_GPIO_LINE_MAX_IRQ) {
		/* Ports A/B/F */
		ep93xx_gpio_int_mask(line);
		ep93xx_gpio_update_int_params(line >> 3);
	}

	v = __raw_readb(ep93xx_chip->data_dir_reg);
	v |= (1 << offset);
	__raw_writeb(v, ep93xx_chip->data_dir_reg);

	local_irq_restore(flags);

	return 0;
}

static int ep93xx_gpio_get(struct gpio_chip *chip, unsigned offset)
{
	struct ep93xx_gpio_chip *ep93xx_chip = to_ep93xx_gpio_chip(chip);

	return !!(__raw_readb(ep93xx_chip->data_reg) & (1 << offset));
}

static void ep93xx_gpio_set(struct gpio_chip *chip, unsigned offset, int val)
{
	struct ep93xx_gpio_chip *ep93xx_chip = to_ep93xx_gpio_chip(chip);
	unsigned long flags;
	u8 v;

	local_irq_save(flags);
	v = __raw_readb(ep93xx_chip->data_reg);
	if (val)
		v |= (1 << offset);
	else
		v &= ~(1 << offset);
	__raw_writeb(v, ep93xx_chip->data_reg);
	local_irq_restore(flags);
}

static void ep93xx_gpio_dbg_show(struct seq_file *s, struct gpio_chip *chip)
{
	struct ep93xx_gpio_chip *ep93xx_chip = to_ep93xx_gpio_chip(chip);
	u8 data_reg, data_dir_reg;
	int i;

	data_reg = __raw_readb(ep93xx_chip->data_reg);
	data_dir_reg = __raw_readb(ep93xx_chip->data_dir_reg);

	for (i = 0; i < chip->ngpio; i++)
		seq_printf(s, "GPIO %s%d: %s %s\n", chip->label, i,
			   (data_reg & (1 << i)) ? "set" : "clear",
			   (data_dir_reg & (1 << i)) ? "out" : "in");
}

#define EP93XX_GPIO_BANK(base_gpio, name, dr, ddr, db)			\
	{								\
		.chip = {						\
			.label		  = name,			\
			.direction_input  = ep93xx_gpio_direction_input, \
			.direction_output = ep93xx_gpio_direction_output, \
			.get		  = ep93xx_gpio_get,		\
			.set		  = ep93xx_gpio_set,		\
			.dbg_show	  = ep93xx_gpio_dbg_show,	\
			.base		  = base_gpio,			\
			.ngpio		  = EP93XX_GPIO_LINE_PER_PORT,	\
		},							\
		.data_reg	= EP93XX_GPIO_REG(dr),			\
		.data_dir_reg	= EP93XX_GPIO_REG(ddr),			\
		.debounce_reg	= EP93XX_GPIO_REG(db),			\
	}

static struct ep93xx_gpio_chip ep93xx_gpio_banks[] = {
	EP93XX_GPIO_BANK(0x00, "A", 0x00, 0x10, 0xA8),
	EP93XX_GPIO_BANK(0x08, "B", 0x04, 0x14, 0xC4),
	EP93XX_GPIO_BANK(0x10, "F", 0x30, 0x34, 0x64),
	EP93XX_GPIO_BANK(0x18, "D", 0x0c, 0x1c, -1),
	EP93XX_GPIO_BANK(0x20, "E", 0x20, 0x24, -1),
	EP93XX_GPIO_BANK(0x28, "C", 0x08, 0x18, -1),
	EP93XX_GPIO_BANK(0x30, "G", 0x38, 0x3c, -1),
	EP93XX_GPIO_BANK(0x38, "H", 0x40, 0x44, -1),
};

void __init ep93xx_gpio_init(void)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(ep93xx_gpio_banks); i++)
		gpiochip_add(&ep93xx_gpio_banks[i].chip);
}

static struct ep93xx_gpio_chip *get_ep93xx_gpio_chip(unsigned int gpio)
{
	int index = gpio / EP93XX_GPIO_LINE_PER_PORT;

	if (index >= ARRAY_SIZE(ep93xx_gpio_banks))
		BUG();

	return &ep93xx_gpio_banks[index];
}

int ep93xx_gpio_direction_get(unsigned int gpio)
{
	struct ep93xx_gpio_chip *ep93xx_chip = get_ep93xx_gpio_chip(gpio);
	unsigned int offset = gpio - ep93xx_chip->chip.base;

	return !!(__raw_readb(ep93xx_chip->data_dir_reg) & (1 << offset));
}
EXPORT_SYMBOL(ep93xx_gpio_direction_get);

int ep93xx_gpio_debounce_set(unsigned int gpio, int arg)
{
	struct ep93xx_gpio_chip *ep93xx_chip;
	unsigned int offset;
	unsigned long flags;
	u8 v;

	if (gpio > EP93XX_GPIO_LINE_MAX_IRQ)
		return -EINVAL;

	ep93xx_chip = get_ep93xx_gpio_chip(gpio);
	offset = gpio - ep93xx_chip->chip.base;

	local_irq_save(flags);
	v = __raw_readb(ep93xx_chip->debounce_reg);
	if (arg)
		v |= (1 << offset);
	else
		v &= ~(1 << offset);
	__raw_writeb(v, ep93xx_chip->debounce_reg);
	local_irq_restore(flags);

	return 0;
}
EXPORT_SYMBOL(ep93xx_gpio_debounce_set);

int ep93xx_gpio_debounce_get(unsigned int gpio)
{
	struct ep93xx_gpio_chip *ep93xx_chip;
	unsigned int offset;

	if (gpio > EP93XX_GPIO_LINE_MAX_IRQ)
		return -EINVAL;

	ep93xx_chip = get_ep93xx_gpio_chip(gpio);
	offset = gpio - ep93xx_chip->chip.base;

	return !!(__raw_readb(ep93xx_chip->debounce_reg) & (1 << offset));
}
EXPORT_SYMBOL(ep93xx_gpio_debounce_get);
