/*
 * drivers/gpio/gpio-ep93xx-sysfs.c
 * GPIO driver using sysfs for ep93xx.
 *
 * Copyright (C) 2009 Atmark Techno, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/platform_device.h>

#include <linux/gpio.h>

#define NR_ATTR 3

extern int ep93xx_gpio_direction_get(unsigned int gpio);
extern int ep93xx_gpio_debounce_set(unsigned int gpio, int arg);
extern int ep93xx_gpio_debounce_get(unsigned int gpio);

struct ep93xx_gpio_port_info {
	struct device_attribute attr_direction;
	struct device_attribute attr_value;
	struct device_attribute attr_debounce;
	struct ep93xx_gpio_pin *pin;
};

struct ep93xx_gpio_drv_info {
	struct attribute_group attr_group;
	int nr_port;
	struct ep93xx_gpio_port_info *pinfos;
};

static ssize_t ep93xx_gpio_direction_show(struct device *dev,
					  struct device_attribute *attr, char *buf)
{
	struct ep93xx_gpio_port_info *port;
	int dir;

	port = container_of(attr, struct ep93xx_gpio_port_info, attr_direction);
	dir = ep93xx_gpio_direction_get(port->pin->gpio);

	return snprintf(buf, PAGE_SIZE, "%d\n", dir);
}

static ssize_t ep93xx_gpio_value_show(struct device *dev,
				      struct device_attribute *attr, char *buf)
{
	struct ep93xx_gpio_port_info *port;
	int data;

	port = container_of(attr, struct ep93xx_gpio_port_info, attr_value);
	data = gpio_get_value(port->pin->gpio);

	return snprintf(buf, PAGE_SIZE, "%d\n", data);
}

static ssize_t ep93xx_gpio_debounce_show(struct device *dev,
					 struct device_attribute *attr, char *buf)
{
	struct ep93xx_gpio_port_info *port;
	int data;

	port = container_of(attr, struct ep93xx_gpio_port_info, attr_debounce);
	data = ep93xx_gpio_debounce_get(port->pin->gpio);

	return snprintf(buf, PAGE_SIZE, "%d\n", data);
}

static ssize_t ep93xx_gpio_direction_store(struct device *dev,
					   struct device_attribute *attr,
					   const char *buf, size_t count)
{
	struct ep93xx_gpio_port_info *port;
	int dir;

	port = container_of(attr, struct ep93xx_gpio_port_info, attr_direction);
	dir = !!simple_strtol(buf, NULL, 0);
	if (dir == 0)
		gpio_direction_input(port->pin->gpio);
	else
		gpio_direction_output(port->pin->gpio, 1);

	return count;
}

static ssize_t ep93xx_gpio_value_store(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf, size_t count)
{
	struct ep93xx_gpio_port_info *port;
	int data;

	port = container_of(attr, struct ep93xx_gpio_port_info, attr_value);
	data = !!simple_strtol(buf, NULL, 0);
	gpio_set_value(port->pin->gpio, data);

	return count;
}

static ssize_t ep93xx_gpio_debounce_store(struct device *dev,
					  struct device_attribute *attr,
					  const char *buf, size_t count)
{
	struct ep93xx_gpio_port_info *port;
	int data;

	port = container_of(attr, struct ep93xx_gpio_port_info, attr_debounce);
	data = !!simple_strtol(buf, NULL, 0);
	ep93xx_gpio_debounce_set(port->pin->gpio, data);

	return count;
}

static int ep93xx_gpio_create_sysfs_node(struct platform_device *pdev,
					 struct ep93xx_gpio_drv_info *dinfo)
{
	struct ep93xx_gpio_platform_data *pdata = pdev->dev.platform_data;
	struct device *dev = &pdev->dev;
	struct attribute **gpio_attrs;
	int ret;
	int i;

	dinfo->pinfos = kzalloc(sizeof(struct ep93xx_gpio_port_info) *
			pdata->num_gpios, GFP_KERNEL);
	if (!dinfo->pinfos) {
		ret = -ENOMEM;
		goto out;
	}

	gpio_attrs = kzalloc(sizeof(struct attribute *) *
			     pdata->num_gpios * NR_ATTR + 1, GFP_KERNEL);
	if (!gpio_attrs) {
		ret = -ENOMEM;
		goto out_kfree1;
	}

#define SET_ATTR(_type, _name)							\
	{									\
		pinfo->attr_##_type.attr.name = _name;				\
		pinfo->attr_##_type.attr.mode = S_IRUGO | S_IWUSR | S_IWGRP;	\
		pinfo->attr_##_type.show      = ep93xx_gpio_##_type##_show;	\
		pinfo->attr_##_type.store     = ep93xx_gpio_##_type##_store;	\
	}

	for (i=0; i < (pdata->num_gpios * NR_ATTR); i += NR_ATTR) {
		struct ep93xx_gpio_port_info *pinfo;

		pinfo = &dinfo->pinfos[i/NR_ATTR];
		pinfo->pin = &pdata->gpios[i/NR_ATTR];

		SET_ATTR(direction, pinfo->pin->direction_name);
		SET_ATTR(value, pinfo->pin->value_name);
		SET_ATTR(debounce, pinfo->pin->debounce_name);

		gpio_attrs[ i ] = (struct attribute *)&pinfo->attr_direction;
		gpio_attrs[i+1] = (struct attribute *)&pinfo->attr_value;
		gpio_attrs[i+2] = (struct attribute *)&pinfo->attr_debounce;

	}
	gpio_attrs[i] = NULL;

	dinfo->nr_port = pdata->num_gpios;
	dinfo->attr_group.name = "ports";
	dinfo->attr_group.attrs = gpio_attrs;

	if (!get_device(dev)) {
		ret = -ENODEV;
		goto out_kfree2;
	}

	ret = sysfs_create_group(&dev->kobj, &dinfo->attr_group);
	if (ret < 0)
		goto out_kfree2;

	put_device(dev);

	return 0;

 out_kfree2:
	kfree(dinfo->attr_group.attrs);
 out_kfree1:
	kfree(dinfo->pinfos);
 out:
	return ret;
}

static int ep93xx_gpio_remove_sysfs_node(struct platform_device *pdev,
					 struct ep93xx_gpio_drv_info *dinfo)
{
	struct device *dev = &pdev->dev;

	sysfs_remove_group(&dev->kobj, &dinfo->attr_group);

	kfree(dinfo->pinfos);
	kfree(dinfo->attr_group.attrs);

	return 0;
}

static int ep93xx_gpio_probe(struct platform_device *pdev)
{
	struct ep93xx_gpio_platform_data *pdata = pdev->dev.platform_data;
	struct ep93xx_gpio_drv_info *dinfo;
	int ret;
	int i;

	dinfo = kzalloc(sizeof(struct ep93xx_gpio_drv_info), GFP_KERNEL);
	if (!dinfo)
		return -ENOMEM;

	for (i = 0; i < pdata->num_gpios; i++) {
		ret = gpio_request(pdata->gpios[i].gpio, pdata->gpios[i].name);
		if (ret) {
			dev_err(&pdev->dev, "unable to request %s\n",
				pdata->gpios[i].name);
			goto gpio_request_fail;
		}
	}

	ret = ep93xx_gpio_create_sysfs_node(pdev, dinfo);
	if (ret < 0)
		goto create_sysfs_node_fail;

	dev_set_drvdata(&pdev->dev, dinfo);

	return 0;

create_sysfs_node_fail:
	kfree(dinfo);

gpio_request_fail:
	while (--i >= 0)
		gpio_free(pdata->gpios[i].gpio);

	return ret;
}

static int ep93xx_gpio_remove(struct platform_device *pdev)
{
	struct ep93xx_gpio_platform_data *pdata = pdev->dev.platform_data;
	struct ep93xx_gpio_drv_info *dinfo = dev_get_drvdata(&pdev->dev);
	int i;

	dev_set_drvdata(&pdev->dev, NULL);

	ep93xx_gpio_remove_sysfs_node(pdev, dinfo);

	for (i = 0; i < pdata->num_gpios; i++)
		gpio_free(pdata->gpios[i].gpio);

	kfree(dinfo);

	return 0;
}

static struct platform_driver ep93xx_gpio_driver = {
	.probe	= ep93xx_gpio_probe,
	.remove	= __devexit_p(ep93xx_gpio_remove),
	.driver	= {
		.name = "ep93xx-gpio",
	},
};

static int __init ep93xx_gpio_init(void)
{
	return platform_driver_register(&ep93xx_gpio_driver);
}

static void __exit ep93xx_gpio_exit(void)
{
	platform_driver_unregister(&ep93xx_gpio_driver);
}

module_init(ep93xx_gpio_init);
module_exit(ep93xx_gpio_exit);

MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_DESCRIPTION("GPIO driver using sysfs for ep93xx");
MODULE_LICENSE("GPL v2");
