/*
 * drivers/ide/arm/ide-ep93xx.c
 *
 * Copyright 2010 Atmark Techno, Inc. All Rights Reserved.
 *
 * Based on the Cirrus Logic ide-ep93xx driver.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *
 * This driver is supported only 200MHz CPU clock speed.
 * To implement in the other CPU clock frequency please adjust 
 * the delay timing parameter in pio_delay[1-3]() macro.  
 *
 */

/* #define DEBUG 1 */
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>
#include <linux/ide.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <asm/arch/dma.h>

#define DRIVER_NAME	"ep93xx-ide"
#define AUTHOR		"Atmark Techno, Inc."
#define DESCRIPTION	"IDE driver"

/*
 * one nop usually equals one cpu cycle (?),
 * so one nop equals 5nsec with a 200MHz clock.
 */
#define delay5ns()	{nop();}
#define delay10ns()	{delay5ns(); delay5ns();}
#define delay25ns()	{delay10ns(); delay10ns(); delay5ns();}
#define delay70ns()	{delay25ns(); delay25ns(); delay10ns(); delay10ns();}

/*
 * PIO Mode0: Delay(S/W) for 70ns
 * PIO Mode1: Delay(S/W) for 50ns
 * PIO Mode2: Delay(S/W) for 30ns
 * PIO Mode3: Delay(S/W) for 30ns
 * PIO Mode4: Delay(S/W) for 25ns
 */
#define pio_delay1()	{delay25ns();}

/*
 * PIO Mode0: Delay(S/W) for 290ns
 * PIO Mode1: Delay(S/W) for 290ns
 * PIO Mode2: Delay(S/W) for 290ns
 * PIO Mode3: Delay(S/W) for 80ns
 * PIO Mode4: Delay(S/W) for 70ns
 */
#define pio_delay2()	{delay70ns();}

/*
 * PIO Mode0: Delay(S/W) for 240ns
 * PIO Mode1: Delay(S/W) for 50ns
 * PIO Mode2: Delay(S/W) for 20ns
 * PIO Mode3: Delay(S/W) for 70ns
 * PIO Mode4: Delay(S/W) for 25ns
 */
#define pio_delay3()	{delay25ns();}

struct ep93xx_dma_info {
	int		channel;

	/*
	 * Global to keep track of the number of entries in the dma buffer
	 * table for each transfer.
	 */
	int		prd_count;
	unsigned int	prd_total;
	unsigned int	prd_returned;

	/*
	 * Set during the DMA callback function to indicate that a ep93xx rev e0
	 * related DMA error occurred 
	 */
	unsigned int	err;

	/*
	 * The count of consective DMA request errors.
	 * { hda reads, hda writes, hdb reads, hdb writes }
	 */
	unsigned long	errors[4];

	/* The particulars of the most recent command that was executed. */
	unsigned long	cmd;
	ide_drive_t	*drive;
	unsigned long	sector;
	unsigned long	nr_sectors;
	unsigned long	cur_retries;

	/*
	 * A garbage word that is used to satify additional DMA requests from the
	 * device.
	 */
	unsigned long	garbage;

	unsigned int	*table;
};

struct ep93xx_ide_info {
	u8		pio_mode;
	u8		mdma_mode;
	u8		udma_mode;

	struct clk	*hclk;
	struct clk	*fclk;

	struct ep93xx_dma_info	*dma;
};

struct ep93xx_ide_private {
	ide_hwif_t	*hwif;
	struct resource	*io, *ctl;
};

#define EP93XX_INFO(drive)	((struct ep93xx_ide_info *)HWIF(drive)->hwif_data)
#define EP93XX_DMA(drive)	(EP93XX_INFO(drive)->dma)

#define RETRIES_BEFORE_PIO_TRANSFER	(ERROR_RESET - 1)
#define FAILURES_BEFORE_DMA_BACKOFF	50

/*
 *  functions to set up the IDE control register and data restister to read
 *  or write 16 bits of data to/from the specified IDE device register.
 *  These functions should only be used when reading/writing data to/from
 *  the data register.
 */
static void ep93xx_ide_outw(unsigned short data, unsigned long addr)
{
	unsigned int tmp;

	/* Write the data out. */
	__raw_writel(data, EP93XX_IDE_DATA_OUT);

	/* Write the address out. */
	tmp = EP93XX_IDE_CTRL_DIORN_OUTPUT | EP93XX_IDE_CTRL_DIOWN_OUTPUT |
		((addr & 7) << EP93XX_IDE_CTRL_DEV_ADDR_OUTPUT_SHIFT) |
		(addr >> 10);
	__raw_writel(tmp, EP93XX_IDE_CTRL);
	pio_delay1();

	/* Toggle the write signal. */
	__raw_writel(tmp & ~EP93XX_IDE_CTRL_DIOWN_OUTPUT, EP93XX_IDE_CTRL);
	pio_delay2();

	__raw_writel(tmp, EP93XX_IDE_CTRL);
	pio_delay3();
}

/*
 * read 16bits from specified ATA register address
 */
static unsigned short ep93xx_ide_inw(unsigned long addr)
{
	unsigned int tmp;

	/* Write the address out. */
	tmp = EP93XX_IDE_CTRL_DIORN_OUTPUT | EP93XX_IDE_CTRL_DIOWN_OUTPUT |
		((addr & 7) << EP93XX_IDE_CTRL_DEV_ADDR_OUTPUT_SHIFT) |
		(addr >> 10);
	__raw_writel(tmp, EP93XX_IDE_CTRL);
	pio_delay1();

	/* Toggle the read signal. */
	__raw_writel(tmp & ~EP93XX_IDE_CTRL_DIORN_OUTPUT, EP93XX_IDE_CTRL);
	pio_delay2();
	__raw_writel(tmp, EP93XX_IDE_CTRL);

	/* Read the data in. */
	tmp = __raw_readl(EP93XX_IDE_DATA_IN);

	return tmp & 0xffff;
}

static void ep93xx_ide_outsw(unsigned long addr, void *buf, unsigned int len)
{
	unsigned short *data = (unsigned short *)buf;

	/* Write out data to the data register 16 bits at a time. */
	while (len--)
		ep93xx_ide_outw(*data++, addr);
}

static void ep93xx_ide_insw(unsigned long addr, void *buf, unsigned int len)
{
	unsigned short *data = (unsigned short *)buf;

	/* Read in data from the data register 16 bits at a time. */
	while (len--)
		*data++ = ep93xx_ide_inw(addr);
}

static void ep93xx_ide_outb(u8 data, unsigned long addr)
{
	ep93xx_ide_outw(data, addr);
}

static void ep93xx_ide_outbsync(ide_drive_t *drive, u8 data, unsigned long addr)
{
	ep93xx_ide_outw(data, addr);
}

static unsigned char ep93xx_ide_inb(unsigned long addr)
{
	return ep93xx_ide_inw(addr);
}

static void ep93xx_input_data(ide_drive_t *drive, struct request *rq,
			      void *buf, unsigned int len)
{
	ide_hwif_t *hwif = HWIF(drive);
	struct ide_io_ports *io_ports = &hwif->io_ports;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	ep93xx_ide_insw(io_ports->data_addr, buf, (len + 1) / 2);
}

static void ep93xx_output_data(ide_drive_t *drive, struct request *rq,
			       void *buf, unsigned int len)
{
	ide_hwif_t *hwif = HWIF(drive);
	struct ide_io_ports *io_ports = &hwif->io_ports;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	ep93xx_ide_outsw(io_ports->data_addr, buf, (len + 1) / 2);
}

static void ep93xx_tf_load(ide_drive_t *drive, ide_task_t *task)
{
	ide_hwif_t *hwif = HWIF(drive);
	struct ide_io_ports *io_ports = &hwif->io_ports;
	struct ide_taskfile *tf = &task->tf;
	u8 HIHI = (task->tf_flags & IDE_TFLAG_LBA48) ? 0xE0 : 0xEF;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	if (task->tf_flags & IDE_TFLAG_FLAGGED)
		HIHI = 0xff;

	ide_set_irq(drive, 1);

	if (task->tf_flags & IDE_TFLAG_OUT_DATA) {
		u16 data = (tf->hob_data << 8) | tf->data;

		ep93xx_ide_outw(data, io_ports->data_addr);
	}

	if (task->tf_flags & IDE_TFLAG_OUT_HOB_FEATURE)
		ep93xx_ide_outb(tf->hob_feature, io_ports->feature_addr);
	if (task->tf_flags & IDE_TFLAG_OUT_HOB_NSECT)
		ep93xx_ide_outb(tf->hob_nsect, io_ports->nsect_addr);
	if (task->tf_flags & IDE_TFLAG_OUT_HOB_LBAL)
		ep93xx_ide_outb(tf->hob_lbal, io_ports->lbal_addr);
	if (task->tf_flags & IDE_TFLAG_OUT_HOB_LBAM)
		ep93xx_ide_outb(tf->hob_lbam, io_ports->lbam_addr);
	if (task->tf_flags & IDE_TFLAG_OUT_HOB_LBAH)
		ep93xx_ide_outb(tf->hob_lbah, io_ports->lbah_addr);

	if (task->tf_flags & IDE_TFLAG_OUT_FEATURE)
		ep93xx_ide_outb(tf->feature, io_ports->feature_addr);
	if (task->tf_flags & IDE_TFLAG_OUT_NSECT)
		ep93xx_ide_outb(tf->nsect, io_ports->nsect_addr);
	if (task->tf_flags & IDE_TFLAG_OUT_LBAL)
		ep93xx_ide_outb(tf->lbal, io_ports->lbal_addr);
	if (task->tf_flags & IDE_TFLAG_OUT_LBAM)
		ep93xx_ide_outb(tf->lbam, io_ports->lbam_addr);
	if (task->tf_flags & IDE_TFLAG_OUT_LBAH)
		ep93xx_ide_outb(tf->lbah, io_ports->lbah_addr);

	if (task->tf_flags & IDE_TFLAG_OUT_DEVICE)
		ep93xx_ide_outb((tf->device & HIHI) | drive->select.all,
				io_ports->device_addr);
}

static void ep93xx_tf_read(ide_drive_t *drive, ide_task_t *task)
{
	ide_hwif_t *hwif = HWIF(drive);
	struct ide_io_ports *io_ports = &hwif->io_ports;
	struct ide_taskfile *tf = &task->tf;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	if (task->tf_flags & IDE_TFLAG_IN_DATA) {
		u16 data = ep93xx_ide_inw(io_ports->data_addr);

		tf->data = data & 0xff;
		tf->hob_data = (data >> 8) & 0xff;
	}

	/* be sure we're looking at the low order bits */
	ep93xx_ide_outb(drive->ctl & ~0x80, io_ports->ctl_addr);

	if (task->tf_flags & IDE_TFLAG_IN_NSECT)
		tf->nsect  = ep93xx_ide_inb(io_ports->nsect_addr);
	if (task->tf_flags & IDE_TFLAG_IN_LBAL)
		tf->lbal   = ep93xx_ide_inb(io_ports->lbal_addr);
	if (task->tf_flags & IDE_TFLAG_IN_LBAM)
		tf->lbam   = ep93xx_ide_inb(io_ports->lbam_addr);
	if (task->tf_flags & IDE_TFLAG_IN_LBAH)
		tf->lbah   = ep93xx_ide_inb(io_ports->lbah_addr);
	if (task->tf_flags & IDE_TFLAG_IN_DEVICE)
		tf->device = ep93xx_ide_inb(io_ports->device_addr);

	if (task->tf_flags & IDE_TFLAG_LBA48) {
		ep93xx_ide_outb(drive->ctl | 0x80, io_ports->ctl_addr);

		if (task->tf_flags & IDE_TFLAG_IN_HOB_FEATURE)
			tf->hob_feature = ep93xx_ide_inb(io_ports->feature_addr);
		if (task->tf_flags & IDE_TFLAG_IN_HOB_NSECT)
			tf->hob_nsect   = ep93xx_ide_inb(io_ports->nsect_addr);
		if (task->tf_flags & IDE_TFLAG_IN_HOB_LBAL)
			tf->hob_lbal    = ep93xx_ide_inb(io_ports->lbal_addr);
		if (task->tf_flags & IDE_TFLAG_IN_HOB_LBAM)
			tf->hob_lbam    = ep93xx_ide_inb(io_ports->lbam_addr);
		if (task->tf_flags & IDE_TFLAG_IN_HOB_LBAH)
			tf->hob_lbah    = ep93xx_ide_inb(io_ports->lbah_addr);
	}
}

static void ep93xx_ide_set_pio_mode(ide_drive_t *drive, const u8 pio)
{
	struct ep93xx_ide_info *info = EP93XX_INFO(drive);
	unsigned long tmp;
	unsigned long hclk, fclk;
	unsigned long period, delay, wst;

	pr_debug("%s: PIO Mode%d\n", DRIVER_NAME, pio);

	fclk = clk_get_rate(info->fclk);

	/*
	 * PIO Mode0: Delay(H/W) for 30ns
	 * PIO Mode1: Delay(H/W) for 20ns
	 * PIO Mode2: Delay(H/W) for 15ns
	 * PIO Mode3: Delay(H/W) for 10ns
	 * PIO Mode4: Delay(H/W) for  5ns
	 */
	switch (pio) {
	case 4:
		/*
		 * In PIO Mode4 delay(S/W) 70ns is the most slowest delay.
		 *
		 * 1 / 200MHz = 5ns
		 * 70ns / 5ns = 14
		 * (5ns / (70ns / (14 - 1))) * 200MHz = 185714286Hz
		 */
		if (fclk < 185714286)
			pr_info("%s: PIO delay(S/W) is too slow."
				"See pio_delay[1-3]() macro.\n", DRIVER_NAME);;
		delay = 5;
		break;
	default:
		pr_warning("%s: %s mode is not supported\n",
			   DRIVER_NAME, ide_xfer_verbose(pio + XFER_PIO_0));
		return;
	}

	hclk = clk_get_rate(info->hclk);
	/*
	 * This gives us the period in nanosecods.
	 *
	 * = 1000000000(ns/s) / HCLK(cycle/s)
	 * = (ns/cycle)
	 */
	period = 1000000000 / hclk;

	wst = delay / period;
	if (delay % period)
		wst++;
	if (wst > (EP93XX_IDE_CONFIG_WAIT_TIME_MASK >>
		   EP93XX_IDE_CONFIG_WAIT_TIME_SHIFT))
		wst = (EP93XX_IDE_CONFIG_WAIT_TIME_MASK >>
		       EP93XX_IDE_CONFIG_WAIT_TIME_SHIFT);

	/* Insure that neither device is selected. -wlg */
	tmp = __raw_readl(EP93XX_IDE_CTRL) |
		(EP93XX_IDE_CTRL_CS0_OUTPUT | EP93XX_IDE_CTRL_CS1_OUTPUT);
	__raw_writel(tmp, EP93XX_IDE_CTRL);

	/* Make sure that MWDMA and UDMA are disabled. */
	__raw_writel(0, EP93XX_IDE_MDMA_OPERATION);
	__raw_writel(0, EP93XX_IDE_UDMA_OPERATION);

	/* Reset the UDMA state machine. */
	__raw_writel(EP93XX_IDE_UDMA_DEBUG_MASK, EP93XX_IDE_UDMA_DEBUG);
	__raw_writel(0, EP93XX_IDE_UDMA_DEBUG);

	/* Enable PIO mode */
	tmp = EP93XX_IDE_CONFIG_IDE_ENABLE | EP93XX_IDE_CONFIG_POLL_IO |
		(pio << EP93XX_IDE_CONFIG_MODE_SHIFT) |
		(wst << EP93XX_IDE_CONFIG_WAIT_TIME_SHIFT);
	__raw_writel(tmp, EP93XX_IDE_CONFIG);

	info->pio_mode = pio;
}

static void ep93xx_ide_set_dma_mode(ide_drive_t *drive, const u8 speed)
{
	struct ep93xx_ide_info *info = EP93XX_INFO(drive);

	switch (speed) {
	case XFER_UDMA_0:
	case XFER_UDMA_1:
	case XFER_UDMA_2:
	case XFER_UDMA_3:
		break;
	default:
		pr_warning("%s: %s mode is not supported\n",
			   DRIVER_NAME, ide_xfer_verbose(speed));
		return;
	}

	info->udma_mode = speed - XFER_UDMA_0;

	pr_debug("%s: %s mode selected\n", DRIVER_NAME,
		 ide_xfer_verbose(speed));
}

static u8 ep93xx_ide_cable_detect(ide_hwif_t *hwif)
{
	if (!strncmp(CONFIG_BLK_DEV_IDE_EP93XX_CABLE_TYPE, "80-", 3))
		return ATA_CBL_PATA80;

	return ATA_CBL_PATA40;
}

static int ep93xx_ide_init_dma(ide_hwif_t *hwif, const struct ide_port_info *d)
{
	struct ep93xx_ide_info *info = hwif->hwif_data;
	struct ep93xx_dma_info *dma;
	unsigned long id;
	int ret;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	id = ep93xx_chip_revision();
	if (id <= EP93XX_SYSCON_CHIP_ID_REVISION_D0) {
		pr_notice("%s: ep93xx chip revision too old for IDE DMA support\n",
			  DRIVER_NAME);
		ret = -EOPNOTSUPP;
		goto failed;
	}

	dma = kmalloc(sizeof(struct ep93xx_dma_info), GFP_KERNEL);
	if (!dma) {
		pr_err("%s: DMA disabled, unable to allocate memory.\n", DRIVER_NAME);
		ret = -ENOMEM;
		goto failed;
	}
	memset(dma, 0, sizeof(struct ep93xx_dma_info));

	dma->table = kmalloc(PRD_ENTRIES * PRD_BYTES, GFP_KERNEL);
	if (!dma->table) {
		pr_err("%s: DMA disabled, unable to allocate memory.\n", DRIVER_NAME);
		ret = -ENOMEM;
		goto free_dma;
	}

	/* Open an instance of the ep93xx dma interface. */
	ret = ep93xx_dma_request(&dma->channel, hwif->name, DMA_IDE);
	if (ret) {
		kfree(dma->table);
		pr_err("%s: DMA disabled, unable to dma request.\n", DRIVER_NAME);
		goto free_table;
	}

	/*
	 * Now that we've got a dma channel allocated, set up the rest
	 * of the dma specific stuff.
	 */
	pr_debug("%s: dma channel allocated: 0x%x, %s  %d \n",
		 DRIVER_NAME, dma->channel, hwif->name, DMA_IDE);

	info->dma = dma;

	return 0;

 free_table:
	kfree(dma->table);
 free_dma:
	kfree(dma);
 failed:
	return ret;
}

static void ep93xx_ide_chipinit(void)
{
	unsigned long tmp;

	/* Insure that neither device is selected. -wlg */
	tmp = __raw_readl(EP93XX_IDE_CTRL) |
		(EP93XX_IDE_CTRL_CS0_OUTPUT | EP93XX_IDE_CTRL_CS1_OUTPUT);
        __raw_writel(tmp, EP93XX_IDE_CTRL);

	/* Make sure that MWDMA and UDMA are disabled. */
	__raw_writel(0, EP93XX_IDE_MDMA_OPERATION);
	__raw_writel(0, EP93XX_IDE_UDMA_OPERATION);

	/* Reset the UDMA state machine. */
	__raw_writel(EP93XX_IDE_UDMA_DEBUG_MASK, EP93XX_IDE_UDMA_DEBUG);
	__raw_writel(0, EP93XX_IDE_UDMA_DEBUG);
}

static void ep93xx_ide_dma_host_set(ide_drive_t *drive, int on)
{
	pr_debug("%s: DMA %s\n", DRIVER_NAME, on ? "ON" : "OFF");
}

/*
 * ep93xx_ide_callback() - cleanup after dma transfer
 *
 * Registered with the ep93xx dma driver and called at the end of the dma
 * interrupt handler, this function processes the dma buffers.
 */
static void ep93xx_ide_callback(ep93xx_dma_int_t dma_int,
				ep93xx_dma_dev_t device, unsigned int user_data)
{
	ide_drive_t *drive = (ide_drive_t *)user_data;
	ide_hwif_t *hwif = HWIF(drive);
	struct ep93xx_dma_info *dma = EP93XX_DMA(drive);
	struct request *rq = HWGROUP(drive)->rq;
	unsigned int src, dest;
	unsigned int temp;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	/*
	 * Retrieve from the dma interface as many used buffers as are
	 * available.
	 */
	while (ep93xx_dma_remove_buffer(dma->channel, &temp) == 0)
		dma->prd_returned++;

	/* add any remaining buffers */
	while (dma->prd_count) {
		if (rq_data_dir(rq) == READ) {
			src = hwif->dma_base;
			dest = hwif->dmatable_cpu[0];
		} else {
			src = hwif->dmatable_cpu[0];
			dest = hwif->dma_base;
		}

		if (ep93xx_dma_add_buffer(dma->channel, src, dest,
					  hwif->dmatable_cpu[1], 0,
					  dma->prd_count) != 0)
			break;
		hwif->dmatable_cpu += 2;
		dma->prd_count--;
	};

	/*
	 * check if all of the buffers have been returned, but the device
	 * still wants more.
	 */
   	if ((dma->prd_returned >= dma->prd_total) &&
	    (__raw_readl(EP93XX_IDE_CTRL) & EP93XX_IDE_CTRL_DMARQ_STATE)) {
		dma->err = 1;
		pr_warning("%s: DMA transfer error detected: "
			   "transfer complete but device wants more\n",
			   DRIVER_NAME);
		if (hwif->sg_dma_direction == PCI_DMA_FROMDEVICE)
			ep93xx_dma_add_buffer(dma->channel, hwif->dma_base,
					      virt_to_dma(NULL, &dma->garbage),
					      4, 0, 0);
		else
			ep93xx_dma_add_buffer(dma->channel,
					      virt_to_dma(NULL, &dma->garbage),
					      hwif->dma_base, 4, 0, 0);
	}

	pr_debug("\ttreturn from callback\n");
}

static int ep93xx_ide_dma_setup(ide_drive_t *drive)
{
	struct ep93xx_dma_info *dma = EP93XX_DMA(drive);
	struct request *rq = HWGROUP(drive)->rq;
	unsigned int flags, i;
	ide_hwif_t *hwif = HWIF(drive);
	unsigned int src, dest;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);
	pr_debug("\t%ld, %ld\n", rq->sector, rq->nr_sectors);

	if (drive->waiting_for_dma) {
		pr_warning("%s: dma already active \n", DRIVER_NAME);
		return 1;
	}

	/* Compute the array index used for this request. */
	i = rq_data_dir(rq) == READ ? 0 : 1;
	i += drive->name[2] == 'a' ? 0 : 2;

	/*
	 * See if this request matches the most recent request, and that the
	 * most recent request ended in error.
	 */
	if (dma->err && (dma->cmd == rq_data_dir(rq)) && (dma->drive == drive) &&
	    (dma->sector == rq->sector) && (dma->nr_sectors == rq->nr_sectors)) {
		dma->cur_retries++;
		if (dma->cur_retries == RETRIES_BEFORE_PIO_TRANSFER) {
			pr_warning("%s: using pio for this request\n", DRIVER_NAME);
			dma->err = 0;
			return 1;
		}
	} else {
		dma->cur_retries = 0;
		dma->cmd = rq_data_dir(rq);
		dma->drive = drive;
		dma->sector = rq->sector;
		dma->nr_sectors = rq->nr_sectors;
	}

	/*
	 * Configure DMA M2M channel flags for a source/destination address
	 * hold, h/w initiated P2M/M2P transfer.
	 */
	flags = rq_data_dir(rq) == READ ?
		(SOURCE_HOLD | TRANSFER_MODE_HW_P2M) :
		(DESTINATION_HOLD | TRANSFER_MODE_HW_M2P);

	if (drive->current_speed & XFER_MW_DMA_0) {
		pr_warning("%s: attempt to setup MDMA\n", DRIVER_NAME);
		return 1;
	} else {
		unsigned long addr;

		flags |= ((rq_data_dir(rq) == READ ?
			   WS_IDE_UDMA_READ : WS_IDE_UDMA_WRITE) <<
			  WAIT_STATES_SHIFT);

		/* UDMA data register address. */
		if (rq_data_dir(rq) == READ)
			addr = EP93XX_IDE_UDMA_DATA_IN;
		else
			addr = EP93XX_IDE_UDMA_DATA_OUT;
		addr = addr - EP93XX_IO_BASE_VIRT + EP93XX_IO_BASE_PHYS;
		hwif->dma_base = virt_to_dma(NULL, addr);
	}

	/* Configure the dma interface for this IDE operation. */
	pr_debug("\tstart setup ep93xx_dma_config\n \n");
	if (ep93xx_dma_config(dma->channel, 0, flags, ep93xx_ide_callback,
			      (unsigned int)drive) != 0) {
		pr_err("%s: ep93xx_ide_dma_setup: ERROR- dma config failed",
		       DRIVER_NAME);
		return 1;
	}
	pr_debug("\tfinish setup ep93xx_dma_config\n\n");

	/* Build the table of dma-able buffers. */
	hwif->dmatable_cpu = dma->table;
	dma->prd_count = ide_build_dmatable(drive, rq);
	if (!dma->prd_count) {
		pr_err("%s: ep93xx_ide_dma_setup: failed to build dma table",
		       DRIVER_NAME);
		return 1;
	}

	dma->prd_total = dma->prd_count;
	dma->prd_returned = 0;

	/* Add buffers to the DMA interface from the dma_table. */
	do {
		if (rq_data_dir(rq) == READ) {
			src = hwif->dma_base;
			dest = hwif->dmatable_cpu[0];
		} else {
			src = hwif->dmatable_cpu[0];
			dest = hwif->dma_base;
		}

		if (ep93xx_dma_add_buffer(dma->channel, src, dest,
					  hwif->dmatable_cpu[1], 0,
					  dma->prd_count) != 0)
			break;
		hwif->dmatable_cpu += 2;
		dma->prd_count--;
	} while (dma->prd_count);

	/* Indicate that we're waiting for dma. */
	drive->waiting_for_dma = 1;

	return 0;
}

/*
 * ep93xx_ide_dma_intr() - handler for disk read/write DMA interrupts
 */
static ide_startstop_t ep93xx_ide_dma_intr(ide_drive_t *drive)
{
	byte stat, dma_stat;
	ide_hwif_t *hwif = HWIF(drive);
	struct ide_io_ports *io_ports = &hwif->io_ports;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	dma_stat = hwif->dma_ops->dma_end(drive);
	stat = hwif->INB(io_ports->status_addr);
	if (OK_STAT(stat, DRIVE_READY, drive->bad_wstat | DRQ_STAT)) {
		if (!dma_stat) {
			struct request *rq = HWGROUP(drive)->rq;
			ide_driver_t *drv =
				*(ide_driver_t **)rq->rq_disk->private_data;
			drv->end_request(drive, 1, rq->nr_sectors);
			return ide_stopped;
		}
		pr_err("%s: dma_intr: bad DMA status (dma_stat=%x)\n",
		       DRIVER_NAME, dma_stat);
	}

	return ide_error(drive, __func__, stat);
}

/*
 * ep93xx_dma_timer_expiry() - handle a DMA timeout
 * 
 * See dma_timer_expiry() in ide-dma.c for comments.
 */
static int ep93xx_idedma_timer_expiry(ide_drive_t *drive)
{
	ide_hwif_t *hwif = HWIF(drive);
	struct ide_io_ports *io_ports = &hwif->io_ports;
	u8 dev_stat = hwif->INB(io_ports->ctl_addr);
	u8 irq_stat = __raw_readl(EP93XX_IDE_CTRL) & EP93XX_IDE_CTRL_INTRQ_STATE;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);
	pr_warning("%s: dma_timer_expiry: dev status == 0x%02x,irq= %d\n",
		   DRIVER_NAME, dev_stat, irq_stat);

	/*
	 * Clear the expiry handler in case we decide to wait more,
	 * next time timer expires it is an error
	 */
	HWGROUP(drive)->expiry = NULL;

	/* If the interrupt is asserted, call the handler. */
	if (irq_stat)
		HWGROUP(drive)->handler(drive);

	/*
	 * Check if the busy bit or the drq bit is set, indicating that
	 * a dma transfer is still active, or the IDE interrupt is asserted.
	 */
	if ((dev_stat & BUSY_STAT) || (dev_stat & DRQ_STAT) || irq_stat)
		return WAIT_CMD;

	/*
	 * the device is not busy and the interrupt is not asserted, so check
	 * if there's an error.
	 */
	if (dev_stat & ERR_STAT) /* ERROR */
		return -1;

	pr_debug("\t(end)\n");

	return 0; /* Unknown status -- reset the bus */
}

static void ep93xx_ide_dma_exec_cmd(ide_drive_t *drive, u8 command)
{
	pr_debug("%s: %s\n", DRIVER_NAME, __func__);
	pr_debug("\tcommand: %x\n", command);

	ide_execute_command(drive, command, &ep93xx_ide_dma_intr, 2 * WAIT_CMD,
			    &ep93xx_idedma_timer_expiry);

	pr_debug("\t(end)\n");
}

static void ep93xx_ide_dma_start(ide_drive_t *drive)
{
	struct ep93xx_ide_info *info = EP93XX_INFO(drive);
	struct ep93xx_dma_info *dma = EP93XX_DMA(drive);
	struct request *rq = HWGROUP(drive)->rq;
	unsigned long tmp;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);
	pr_debug("\t%s\n", rq_data_dir(rq) == READ ? "read" : "write");

	/* the transfer is not bad yet. */
	dma->err = 0;

	/* Insure that neither device is selected. -wlg */
	tmp = __raw_readl(EP93XX_IDE_CTRL) |
		(EP93XX_IDE_CTRL_CS0_OUTPUT | EP93XX_IDE_CTRL_CS1_OUTPUT);
	__raw_writel(tmp, EP93XX_IDE_CTRL);

	/* Enable UDMA mode */
	tmp = EP93XX_IDE_CONFIG_UDMA | EP93XX_IDE_CONFIG_IDE_ENABLE |
		(info->udma_mode << EP93XX_IDE_CONFIG_MODE_SHIFT);
	__raw_writel(tmp, EP93XX_IDE_CONFIG);

	pr_debug("%s: UDMA Mode%d\n", DRIVER_NAME, info->udma_mode);

	if (rq_data_dir(rq) != READ)
		tmp = EP93XX_IDE_UDMA_OPERATION_WRITE;
	__raw_writel(tmp, EP93XX_IDE_UDMA_OPERATION);

	tmp = __raw_readl(EP93XX_IDE_UDMA_OPERATION) |
		EP93XX_IDE_UDMA_OPERATION_ENABLE;
	__raw_writel(tmp, EP93XX_IDE_UDMA_OPERATION);

	/* start the dma transfer. */
	ep93xx_dma_start(dma->channel, 1, NULL);

	pr_debug("\t%s end\n", __func__);
}

static inline unsigned int ep93xx_get_head_pointer(unsigned int status)
{
	return status & EP93XX_IDE_UDMA_BUFFER_STATUS_HEADPOINTER_MASK;
}

static inline unsigned int ep93xx_get_tail_pointer(unsigned int status)
{
	return (status & EP93XX_IDE_UDMA_BUFFER_STATUS_TAILPOINTER_MASK) >>
		EP93XX_IDE_UDMA_BUFFER_STATUS_TAILPOINTER_SHIFT;
}

static int ep93xx_ide_dma_end(ide_drive_t *drive)
{
	struct ep93xx_ide_info *info = EP93XX_INFO(drive);
	struct ep93xx_dma_info *dma = EP93XX_DMA(drive);
	struct request *rq = HWGROUP(drive)->rq;
	int status;

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	/*
	 * ep93xx rev e0 error detection.
	 * See if there is any data left in UDMA FIFOs.  For a read, first wait
	 * until either the DMA is done or the FIFO is empty.  If there is any
	 * residual data in the FIFO, there was an error in the transfer.
	 */
	if (rq_data_dir(rq) == READ) {
		do {
			status = __raw_readl(EP93XX_IDE_UDMA_READ_BUFFER_STATUS);
		} while (!ep93xx_dma_is_done(dma->channel) &&
			 (ep93xx_get_head_pointer(status) !=
			  ep93xx_get_tail_pointer(status)));
		udelay(1);
	} else
		status = __raw_readl(EP93XX_IDE_UDMA_WRITE_BUFFER_STATUS);

	if (ep93xx_get_head_pointer(status) != ep93xx_get_tail_pointer(status)) {
		pr_warning("%s: DMA transfer error detected: "
			   "fifo residual data\n", DRIVER_NAME);
		dma->err = 1;
	}

	/*
	 * ep93xx rev e0 error detection.
	 * See if all of the buffers were returned by the DMA driver.  If not,
	 * then a transfer error has occurred.
	 */
	if (!ep93xx_dma_is_done(dma->channel)) {
		pr_warning("%s: DMA transfer error detected: "
			   "transfer not complete\n", DRIVER_NAME);
		dma->err = 1;
	}

	ep93xx_dma_pause(dma->channel, 1, 0);
	ep93xx_dma_flush(dma->channel);

	ep93xx_ide_set_pio_mode(drive, info->pio_mode);

	/* Indicate there's no dma transfer currently in progress. */
	drive->waiting_for_dma = 0;

	/* compute the drive array index used for this request. */
	status = rq_data_dir(rq) == READ ? 0 : 1;
	status += drive->name[2] == 'a' ? 0 : 2;

	if (dma->err == 0)
		/* reset the consecutive error counter. */
		dma->errors[status] = 0;
	else {
		dma->errors[status]++;
		if (dma->errors[status] == FAILURES_BEFORE_DMA_BACKOFF) {
			pr_warning("%s: lowering UDMA mode\n", DRIVER_NAME);

			switch (drive->current_speed) {
			case XFER_UDMA_3:
			case XFER_UDMA_2:
			case XFER_UDMA_1:
				ide_config_drive_speed(drive,
						       drive->current_speed - 1);
				break;
			case XFER_UDMA_0:
				HWIF(drive)->dma_ops->dma_host_set(drive, 0);
				ide_config_drive_speed(drive, XFER_PIO_4);
				break;
			}

			/* reset the consecutive error counters */
			dma->errors[status] = 0;
			dma->errors[status ^ 1] = 0;
		}

		return 1;
	}

	return 0;
}

static int ep93xx_ide_dma_test_irq(ide_drive_t *drive)
{
	if (!drive->waiting_for_dma)
		pr_warning("%s called while not waiting\n", DRIVER_NAME);

	return __raw_readl(EP93XX_IDE_CTRL) & EP93XX_IDE_CTRL_INTRQ_STATE;
}

static void ep93xx_ide_dma_timeout(ide_drive_t *drive)
{
	ide_hwif_t *hwif = HWIF(drive);
	const struct ide_dma_ops *dma_ops = hwif->dma_ops;

	pr_err("%s: timeout waiting for DMA\n", DRIVER_NAME);

	if (dma_ops->dma_test_irq(drive))
		return;

	dma_ops->dma_end(drive);
}

static void ep93xx_ide_dma_lost_irq(ide_drive_t *drive)
{
	pr_debug("%s: %s\n", DRIVER_NAME, __func__);
}

static const struct ide_port_ops ep93xx_ide_ports_ops = {
	.set_pio_mode		= ep93xx_ide_set_pio_mode,
	.set_dma_mode		= ep93xx_ide_set_dma_mode,
	.cable_detect		= ep93xx_ide_cable_detect,
};

static const struct ide_dma_ops ep93xx_ide_dma_ops = {
	.dma_host_set		= ep93xx_ide_dma_host_set,
	.dma_setup		= ep93xx_ide_dma_setup,
	.dma_exec_cmd		= ep93xx_ide_dma_exec_cmd,
	.dma_start		= ep93xx_ide_dma_start,
	.dma_end		= ep93xx_ide_dma_end,
	.dma_test_irq		= ep93xx_ide_dma_test_irq,
	.dma_lost_irq		= ep93xx_ide_dma_lost_irq,
	.dma_timeout		= ep93xx_ide_dma_timeout,
};

static struct ide_port_info __devinitdata ep93xx_ide_port_info = {
	.name			= DRIVER_NAME,
	.init_dma		= ep93xx_ide_init_dma,
	.port_ops		= &ep93xx_ide_ports_ops,
	.dma_ops		= &ep93xx_ide_dma_ops,
	.host_flags		= IDE_HFLAG_MMIO,
	.pio_mask		= ATA_PIO4,
};

static int __devinit ep93xx_ide_probe(struct platform_device *pdev)
{
	struct ep93xx_ide_info *info;
	struct ep93xx_ide_private *priv;
	struct resource *io, *ctl;
	int irq;
	ide_hwif_t *hwif;
	hw_regs_t hw;
	int err;
	unsigned int id;
	struct clk *hclk, *fclk;
	unsigned long rate;
	u8 idx[4] = {0xff, 0xff, 0xff, 0xff};

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	io = platform_get_resource(pdev, IORESOURCE_IO, 0);
	if (io == NULL) {
		pr_err("%s: failed to get io region resource\n", DRIVER_NAME);
		err = -ENODEV;
		goto failed;
	}

	ctl = platform_get_resource(pdev, IORESOURCE_IO, 1);
	if (ctl == NULL) {
		pr_err("%s: failed to get ctl region resource\n", DRIVER_NAME);
		err = -ENODEV;
		goto failed;
	}

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		pr_err("%s: failed to get IRQ resource\n", DRIVER_NAME);
		err = -ENODEV;
		goto failed;
	}

	hclk = clk_get(&pdev->dev, "hclk");
	if (IS_ERR(hclk)) {
		pr_err("%s: could not get HCLK\n", DRIVER_NAME);
		err = -ENOENT;
		goto failed;
	}

	fclk = clk_get(&pdev->dev, "fclk");
	if (IS_ERR(fclk)) {
		pr_err("%s: could not get FCLK\n", DRIVER_NAME);
		err = -ENOENT;
		goto failed;
	}

	if (!request_region(io->start, io->end - io->start + 1, DRIVER_NAME)) {
		pr_err("%s: I/O resource 0x%x not free.\n",
		       DRIVER_NAME, io->start);
		err = -EBUSY;
		goto failed;
	}

	if (!request_region(ctl->start, ctl->end - ctl->start + 1, DRIVER_NAME)) {
		pr_err("%s: I/O resource 0x%x not free.\n",
		       DRIVER_NAME, ctl->start);
		err = -EBUSY;
		goto free_io;
	}

	ep93xx_ide_chipinit();

	memset(&hw, 0, sizeof(hw));
	ide_std_init_ports(&hw, io->start, ctl->start);
	hw.irq = irq;
	hw.chipset = ide_unknown;
	hw.dev = &pdev->dev;

	hwif = ide_find_port();
	if (!hwif) {
		pr_err("%s: register fail\n", DRIVER_NAME);
		err = -ENODEV;
		goto free_ctl;
	}

	ide_init_port_data(hwif, hwif->index);
	ide_init_port_hw(hwif, &hw);

	hwif->tf_load		= ep93xx_tf_load;
	hwif->tf_read		= ep93xx_tf_read;
	hwif->input_data	= ep93xx_input_data;
	hwif->output_data	= ep93xx_output_data;
	hwif->OUTB		= ep93xx_ide_outb;
	hwif->OUTBSYNC		= ep93xx_ide_outbsync;
	hwif->INB		= ep93xx_ide_inb;

	priv = kmalloc(sizeof(struct ep93xx_ide_private), GFP_KERNEL);
	if (!priv) {
		pr_err("%s: unable to allocate memory.\n", DRIVER_NAME);
		err = -ENOMEM;
		goto free_ctl;
	}
	platform_set_drvdata(pdev, priv);
	priv->hwif = hwif;
	priv->io = io;
	priv->ctl = ctl;

	info = kmalloc(sizeof(struct ep93xx_ide_info), GFP_KERNEL);
	if (!info) {
		pr_err("%s: unable to allocate memory.\n", DRIVER_NAME);
		err = -ENOMEM;
		goto free_priv;
	}
	hwif->hwif_data = info;
	info->hclk = hclk;
	info->fclk = fclk;

	rate = clk_get_rate(clk_get(&pdev->dev, "fclk"));
	if (rate > 200000000) {
		pr_err("%s: PIO delay(S/W) is too short."
		       "See pio_delay[1-3]() macro.\n", DRIVER_NAME);
		goto free_info;
	}

	idx[0] = hwif->index;

	/*  
	 * On Rev D1/E0 the maximum DMA mode is 2. 
	 * On Rev E1 and later the maximum UDMA mode is 4.
	 * At the moment, we're only allowing up to mode 3.
	 */
	id = ep93xx_chip_revision();
	if (id < EP93XX_SYSCON_CHIP_ID_REVISION_E1)
		ep93xx_ide_port_info.udma_mask = ATA_UDMA2;
	else
		ep93xx_ide_port_info.udma_mask = ATA_UDMA3;

	ide_device_add(idx, &ep93xx_ide_port_info);

	return 0;

 free_info:
	kfree(info);
 free_priv:
	kfree(priv);
	platform_set_drvdata(pdev, NULL);
 free_ctl:
	release_region(ctl->start, ctl->end - ctl->start + 1);
 free_io:
	release_region(io->start, io->end - io->start + 1);
 failed:
	return err;
}

static int __devinit ep93xx_ide_remove(struct platform_device *pdev)
{
	struct ep93xx_ide_private *priv = platform_get_drvdata(pdev);

	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	if (priv) {
		if (priv->hwif) {
			ide_hwif_t *hwif = priv->hwif;

			if (hwif->hwif_data) {
				struct ep93xx_ide_info *info = hwif->hwif_data;

				if (info->dma) {
					struct ep93xx_dma_info *dma = info->dma;
					kfree(dma->table);
					kfree(dma);
				}
				kfree(info);
			}
			ide_unregister(hwif);
		}
		if (priv->ctl)
			release_region(priv->ctl->start,
				       priv->ctl->end - priv->ctl->start + 1);
		if (priv->io)
			release_region(priv->io->start,
				       priv->io->end - priv->io->start + 1);
		kfree(priv);
	}

	platform_set_drvdata(pdev, NULL);

	return 0;
}

static struct platform_driver ep93xx_ide_driver = {
	.driver		= {
		.name	= DRIVER_NAME,
	},
	.probe		= ep93xx_ide_probe,
	.remove		= __devexit_p(ep93xx_ide_remove)
};

static int __init ep93xx_ide_init(void)
{
	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	return platform_driver_register(&ep93xx_ide_driver);
}
module_init(ep93xx_ide_init);

static void __exit ep93xx_ide_exit(void)
{
	pr_debug("%s: %s\n", DRIVER_NAME, __func__);

	platform_driver_unregister(&ep93xx_ide_driver);
}
module_exit(ep93xx_ide_exit);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE("GPL");
