/*
 *  Copyright 2009 Atmark Techno, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/platform_device.h>
#include <linux/leds.h>

#include <asm/arch/gpio.h>

struct armadillo2x0_led_info {
	struct led_classdev cdev;
	struct gpio_led *desc;
};

static void
armadillo2x0_led_set(struct led_classdev *cdev,
		     enum led_brightness value)
{
	struct armadillo2x0_led_info *info = dev_get_drvdata(cdev->dev);

	gpio_direction_output(info->desc->gpio, 1);
	gpio_set_value(info->desc->gpio, !!value);
}

static enum led_brightness
armadillo2x0_led_get(struct led_classdev *cdev)
{
	struct armadillo2x0_led_info *info = dev_get_drvdata(cdev->dev);

	return gpio_get_value(info->desc->gpio) ? LED_FULL : LED_OFF;
}

static int
armadillo2x0_led_probe(struct platform_device *pdev)
{
	struct gpio_led_platform_data *priv = pdev->dev.platform_data;
	struct armadillo2x0_led_info *info;
	int ret;
	int i;

	info = kzalloc(sizeof(struct armadillo2x0_led_info) * priv->num_leds,
		       GFP_KERNEL);
	if (!info)
		return -ENOMEM;

	for (i = 0; i < priv->num_leds; i++) {
		gpio_request(priv->leds[i].gpio, priv->leds[i].name);

		info[i].desc = &priv->leds[i];
		info[i].cdev.name = priv->leds[i].name;
		info[i].cdev.brightness_set = armadillo2x0_led_set;
		info[i].cdev.brightness_get = armadillo2x0_led_get;
		info[i].cdev.default_trigger = "timer";

		ret = led_classdev_register(&pdev->dev, &info[i].cdev);
		if (ret < 0) {
			while (i-- >= 0)
				led_classdev_unregister(&info[i].cdev);

			kfree(info);
			return ret;
		}
	}

	platform_set_drvdata(pdev, info);

	return 0;
}

static int
armadillo2x0_led_remove(struct platform_device *pdev)
{
	struct gpio_led_platform_data *priv = pdev->dev.platform_data;
	struct armadillo2x0_led_info *info = platform_get_drvdata(pdev);
	int i;

	platform_set_drvdata(pdev, NULL);

	for (i = 0; i < priv->num_leds; i++) {
		gpio_free(priv->leds[i].gpio);
		led_classdev_unregister(&info[i].cdev);
	}

	kfree(info);

	return 0;
}

static struct platform_driver armadillo2x0_led_driver = {
	.probe	= armadillo2x0_led_probe,
	.remove	= __devexit_p(armadillo2x0_led_remove),
	.driver	= {
		.name	= "armadillo2x0-led",
	},
};

static int __init
armadillo2x0_led_init(void)
{
	return platform_driver_register(&armadillo2x0_led_driver);
}

static void __exit
armadillo2x0_led_exit(void)
{
	platform_driver_unregister(&armadillo2x0_led_driver);
}

module_init(armadillo2x0_led_init);
module_exit(armadillo2x0_led_exit);

MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_DESCRIPTION("Armadillo-2x0 LED driver");
MODULE_LICENSE("GPL v2");
